"""Base lifecycle hooks/methods. This is *not* a public facing API -- see api.py for classes to extend.
This contains two sets of components:

1. Hooks/methods -- these are classes that customizes Hamilton's execution. They are called at specific
points in Hamilton's execution, and can be used to customize performance. There are specific rules about hooks
and methods
    - Methods can not (currently) be layered. This is because they replace a component of Hamilton's execution
    - Hooks can be layered. Multiple of the same hooks can be called at any given point.
2. Auxiliary tooling to register/manage hooks
    - LifecycleAdapterSet -- this is a class that manages a set of lifecycle adapters. It allows us to call
    all the lifecycle hooks/methods in a given set, and to determine if a given hook/method is implemented.
    - lifecycle -- this is a decorator container that allows us to register hooks/methods. It is used as follows:

To implement a new method/hook type:
1. Create a class that has a single method (see below for examples)
2. Decorate the class with the lifecycle decorator, passing in the name of the method/hook. This must correspond to a method on the class.
3. Add to the LifecycleAdapter type
4. Call out to the hook at different points in the lifecycle

Note that you can have one async hook/method and one sync hook/method in the same class. Some hooks/methods
are coupled to certain execution contexts. While they all live here for now, we could easily envision moving them
externally.

To build an implementation of a hook/method, all you have to do is extend any number of classes.
See api.py for implementations.
"""

import abc
import asyncio
import collections
import dataclasses
import inspect
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

from hamilton import htypes

# We need this because of a (required) circular reference
# Graph depends on lifecycle_base, as it uses the lifecycle hooks
# lifecycle_base has elements that use type-hinting with the FunctionGraph object
# This is OK -- in a *real* compiled language this wouldn't be an issue
# (you'd have header types in C, and java should be smart enough). Given that we're using
# python, which (our usage of) leans type-hinting trigger-happy, this will suffice.
if TYPE_CHECKING:
    from hamilton import graph, node

# All of these are internal APIs. Specifically, structure required to manage a set of
# hooks/methods/validators that we will likely expand. We store them in constants (rather than, say, a more complex single object)
# as it is a clear, simple way to manage the metadata. This allows us to track the registered hooks/methods/validators.

# A set of registered hooks -- each one refers to a string
REGISTERED_SYNC_HOOKS: Set[str] = set()
REGISTERED_ASYNC_HOOKS: Set[str] = set()

# A set of registered methods -- each one refers to a string, which is the name of the metho
REGISTERED_SYNC_METHODS: Set[str] = set()
REGISTERED_ASYNC_METHODS: Set[str] = set()

# A set of registered validators -- these have attached Exception data to them
# Note we do not curently have async validators -- see no need now
REGISTERED_SYNC_VALIDATORS: Set[str] = set()

# constants to refer to internally for hooks
SYNC_HOOK = "hooks"
ASYNC_HOOK = "async_hooks"

# constants to refer to internally for methods
SYNC_METHOD = "methods"
ASYNC_METHOD = "async_methods"

# constants to refer to internally for validators
SYNC_VALIDATOR = "validators"


@dataclasses.dataclass
class ValidationResult:
    success: bool
    error: Optional[str]
    validator: object  # validator so we can make the error message more friendly


class ValidationException(Exception):
    pass


class InvalidLifecycleAdapter(Exception):
    """Container exception to indicate that a lifecycle adapter is invalid."""

    pass


def validate_lifecycle_adapter_function(
    fn: Callable, returns_value: bool, return_type: Optional[Type] = None
):
    """Validates that a function has arguments that are keyword-only,
    and either does or does not return a value, depending on the value of returns_value.

    :param fn: The function to validate
    :param returns_value: Whether the function should return a value or not
    """
    sig = inspect.signature(fn)
    if returns_value:
        if sig.return_annotation is inspect.Signature.empty:
            raise InvalidLifecycleAdapter(
                f"Lifecycle methods must return a value, but {fn} does not have a return annotation."
            )
        if return_type is not None and not htypes.custom_subclass_check(
            sig.return_annotation, return_type
        ):
            raise InvalidLifecycleAdapter(
                f"Lifecycle methods must return a value of type {return_type}, "
                f"but {fn} has a return annotation of "
                f"type {sig.return_annotation}."
            )
    if not returns_value and sig.return_annotation is not inspect.Signature.empty:
        raise InvalidLifecycleAdapter(
            f"Lifecycle hooks/validators must not return a value, but {fn} has a return annotation."
        )
    for param in sig.parameters.values():
        if param.kind != inspect.Parameter.KEYWORD_ONLY and param.name != "self":
            raise InvalidLifecycleAdapter(
                f"Lifecycle methods/hooks can only have keyword-only arguments. "
                f"Method/hook {fn} has argument {param} that is not keyword-only."
            )


def validate_hook_fn(fn: Callable):
    """Validates that a function forms a valid hook. This means:
    1. Function returns nothing
    2. Function must consist of only kwarg-only arguments

    :param fn: The function to validate
    :raises InvalidLifecycleAdapter: If the function is not a valid hook
    """
    validate_lifecycle_adapter_function(fn, returns_value=False)


def validate_method_fn(fn: Callable):
    """Validates that a function forms a valid method. This means:
    1. Function returns a value
    2. Functions must consist of only kwarg-only arguments

    :param fn: The function to validate
    :raises InvalidLifecycleAdapter: If the function is not a valid method
    """
    validate_lifecycle_adapter_function(fn, returns_value=True)


def validate_validator_fn(fn: Callable):
    """Ensures that a function forms a registerable "validator". These are currently the same rules as "hooks".
    While they should also raise an exception, that is not possible to express in the type annotation.

    :param fn: Function to validate
    :raises InvalidLifecycleAdapter: If the function is not a valid validator
    """
    if inspect.iscoroutinefunction(fn):
        raise InvalidLifecycleAdapter(
            f"Lifecycle validators must (so far) be synchronous, "
            f"but {fn} is an async function. "
        )
    validate_lifecycle_adapter_function(fn, returns_value=True)


class lifecycle:
    """Container class for decorators to register hooks/methods.
    This is just a container so it looks clean (`@lifecycle.base_hook(...)`), but we could easily move it out.
    What do these decorators do?
      1. We decorate a class with a method/hook/validator call
      2. This implies that there exists a function by that name
      3. We validate that that function has an appropriate signature
      4. We store this in the appropriate registry (see the constants above)
    Then, when we want to perform a hook/method/validator, we can ask the AdapterLifecycleSet to do so.
    It crawls up the MRO, looking to see which classes are in the registry, then elects which functions to run.
    See LifecycleAdapterSet for more information.
    """

    @classmethod
    def base_hook(cls, fn_name: str):
        """Hooks get called at distinct stages of Hamilton's execution.
        These can be layered together, and potentially coupled to other hooks.

        :param fn_name: Name of the function that will reside in the class we're decorating
        """

        def decorator(clazz):
            fn = getattr(clazz, fn_name, None)
            if fn is None:
                raise ValueError(
                    f"Class {clazz} does not have a method {fn_name}, but is "
                    f'decorated with @lifecycle.base_hook("{fn_name}"). The parameter '
                    f"to @lifecycle.base_hook must be the name "
                    f"of a method on the class."
                )
            validate_hook_fn(fn)
            if inspect.iscoroutinefunction(fn):
                setattr(clazz, ASYNC_HOOK, fn_name)
                REGISTERED_ASYNC_HOOKS.add(fn_name)
            else:
                setattr(clazz, SYNC_HOOK, fn_name)
                REGISTERED_SYNC_HOOKS.add(fn_name)
            return clazz

        return decorator

    @classmethod
    def base_method(cls, fn_name: str):
        """Methods replace the default behavior of Hamilton at a given stage.
        Thus they can only be called once, and not layered. TODO -- determine
        how to allow multiple/have precedence for custom behavior.

        :param fn_name: Name of the function in the class we're registering.
        """

        def decorator(clazz):
            fn = getattr(clazz, fn_name, None)
            if fn is None:
                raise ValueError(
                    f"Class {clazz} does not have a method {fn_name}, but is "
                    f'decorated with @lifecycle.base_hook("{fn_name}"). The parameter '
                    f"to @lifecycle.base_hook must be the name "
                    f"of a method on the class."
                )
            validate_method_fn(fn)
            if inspect.iscoroutinefunction(fn):
                setattr(clazz, ASYNC_METHOD, fn_name)
                REGISTERED_ASYNC_METHODS.add(fn_name)
            else:
                setattr(clazz, SYNC_METHOD, fn_name)
                REGISTERED_SYNC_METHODS.add(fn_name)
            return clazz

        return decorator

    @classmethod
    def base_validator(cls, fn_name: str):
        """Validators are hooks that return a validation result (tuple[success: bool, message: Optional[str]]).
        They provide custom validation logic that runs statically (before the DAG), rather than dynamically (during the DAG run),
        and multiple can be layered together.

        :param fn_name: Name of the function in the class we're registering.
        """

        def decorator(clazz):
            fn = getattr(clazz, fn_name, None)
            if fn is None:
                raise ValueError(
                    f"Class {clazz} does not have a method {fn_name}, but is "
                    f'decorated with @lifecycle.base_validator("{fn_name}"). The parameter '
                    f"to @lifecycle.base_hook must be the name "
                    f"of a method on the class."
                )
            validate_validator_fn(fn)
            setattr(clazz, SYNC_VALIDATOR, fn_name)
            REGISTERED_SYNC_VALIDATORS.add(fn_name)
            return clazz

        return decorator


@lifecycle.base_hook("pre_do_anything")
class BasePreDoAnythingHook(abc.ABC):
    @abc.abstractmethod
    def pre_do_anything(self):
        """Synchronous hook that gets called before doing anything, in the constructor of the driver."""
        pass


@lifecycle.base_method("do_check_edge_types_match")
class BaseDoCheckEdgeTypesMatch(abc.ABC):
    @abc.abstractmethod
    def do_check_edge_types_match(self, *, type_from: type, type_to: type) -> bool:
        """Method that checks whether two types are equivalent. This is used when the function graph is being created.

        :param type_from: The type of the node that is the source of the edge.
        :param type_to: The type of the node that is the destination of the edge.
        :return bool: Whether or not they are equivalent
        """
        pass


@lifecycle.base_method("do_validate_input")
class BaseDoValidateInput(abc.ABC):
    @abc.abstractmethod
    def do_validate_input(self, *, node_type: type, input_value: Any) -> bool:
        """Method that an input value maches an expected type.

        :param node_type:  The type of the node.
        :param input_value:  The value that we want to validate.
        :return: Whether or not the input value matches the expected type.
        """
        pass


@lifecycle.base_validator("validate_node")
class BaseValidateNode(abc.ABC):
    @abc.abstractmethod
    def validate_node(self, *, created_node: "node.Node") -> Tuple[bool, Optional[Exception]]:
        """Validates a node. This will raise an InvalidNodeException
        if the node is invalid.

        :param created_node: Node that was created.
        :raises InvalidNodeException: If the node is invalid.
        """
        pass


@lifecycle.base_validator("validate_graph")
class BaseValidateGraph(abc.ABC):
    @abc.abstractmethod
    def validate_graph(
        self,
        *,
        graph: "graph.FunctionGraph",
        modules: List[ModuleType],
        config: Dict[str, Any],
    ) -> Tuple[bool, Optional[str]]:
        """Validates the graph. This will raise an InvalidNodeException

        :param graph: Graph that has been constructed.
        :param modules: Modules passed into the graph
        :param config: Config passed into the graph
        :return: A (is_valid, error_message) tuple
        """


@lifecycle.base_hook("post_graph_construct")
class BasePostGraphConstruct(abc.ABC):
    @abc.abstractmethod
    def post_graph_construct(
        self,
        *,
        graph: "graph.FunctionGraph",
        modules: List[ModuleType],
        config: Dict[str, Any],
    ):
        """Hooks that is called after the graph is constructed.

        :param graph: Graph that has been constructed.
        :param modules: Modules passed into the graph
        :param config: Config passed into the graph
        """
        pass


@lifecycle.base_hook("post_graph_construct")
class BasePostGraphConstructAsync(abc.ABC):
    @abc.abstractmethod
    async def post_graph_construct(
        self,
        *,
        graph: "graph.FunctionGraph",
        modules: List[ModuleType],
        config: Dict[str, Any],
    ):
        """Asynchronous hook that is called after the graph is constructed.

        :param graph: Graph that has been constructed.
        :param modules: Modules passed into the graph
        :param config: Config passed into the graph
        """
        pass


@lifecycle.base_hook("pre_graph_execute")
class BasePreGraphExecute(abc.ABC):
    @abc.abstractmethod
    def pre_graph_execute(
        self,
        *,
        run_id: str,
        graph: "graph.FunctionGraph",
        final_vars: List[str],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        """Hook that is called immediately prior to graph execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param graph:  Graph that is being executed
        :param final_vars: Variables we are extracting from the graph
        :param inputs: Inputs to the graph
        :param overrides: Overrides to graph execution
        """
        pass


@lifecycle.base_hook("pre_graph_execute")
class BasePreGraphExecuteAsync(abc.ABC):
    @abc.abstractmethod
    async def pre_graph_execute(
        self,
        *,
        run_id: str,
        graph: "graph.FunctionGraph",
        final_vars: List[str],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        """Asynchronous Hook that is called immediately prior to graph execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param graph:  Graph that is being executed
        :param final_vars: Variables we are extracting from the graph
        :param inputs: Inputs to the graph
        :param overrides: Overrides to graph execution
        """
        pass


@lifecycle.base_hook("pre_task_execute")
class BasePreTaskExecute(abc.ABC):
    @abc.abstractmethod
    def pre_task_execute(
        self,
        *,
        run_id: str,
        task_id: str,
        nodes: List["node.Node"],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        """Hook that is called immediately prior to task execution. Note that this is only useful in dynamic
        execution, although we reserve the right to add this back into the standard hamilton execution pattern.

        :param run_id: ID of the run, unique in scope of the driver.
        :param task_id: ID of the task, unique in scope of the driver.
        :param nodes: Nodes that are being executed
        :param inputs: Inputs to the task
        :param overrides: Overrides to task execution
        """
        pass


@lifecycle.base_hook("pre_task_execute")
class BasePreTaskExecuteAsync(abc.ABC):
    @abc.abstractmethod
    async def pre_task_execute(
        self,
        *,
        run_id: str,
        task_id: str,
        nodes: List["node.Node"],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        """Hook that is called immediately prior to task execution. Note that this is only useful in dynamic
        execution, although we reserve the right to add this back into the standard hamilton execution pattern.

        :param run_id: ID of the run, unique in scope of the driver.
        :param task_id: ID of the task, unique in scope of the driver.
        :param nodes: Nodes that are being executed
        :param inputs: Inputs to the task
        :param overrides: Overrides to task execution
        """
        pass


@lifecycle.base_hook("pre_node_execute")
class BasePreNodeExecute(abc.ABC):
    @abc.abstractmethod
    def pre_node_execute(
        self,
        *,
        run_id: str,
        node_: "node.Node",
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
    ):
        """Hook that is called immediately prior to node execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param node_: Node that is being executed
        :param kwargs: Keyword arguments that are being passed into the node
        :param task_id: ID of the task, defaults to None if not in a task setting
        """
        pass


@lifecycle.base_hook("pre_node_execute")
class BasePreNodeExecuteAsync(abc.ABC):
    @abc.abstractmethod
    async def pre_node_execute(
        self,
        *,
        run_id: str,
        node_: "node.Node",
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
    ):
        """Asynchronous hook that is called immediately prior to node execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param node_: Node that is being executed
        :param kwargs: Keyword arguments that are being passed into the node
        :param task_id: ID of the task, defaults to None if not in a task setting
        """
        pass


@lifecycle.base_method("do_node_execute")
class BaseDoNodeExecute(abc.ABC):
    @abc.abstractmethod
    def do_node_execute(
        self,
        *,
        run_id: str,
        node_: "node.Node",
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
    ) -> Any:
        """Method that is called to implement node execution. This can replace the execution of a node
        with something all together, augment it, or delegate it.

        :param run_id: ID of the run, unique in scope of the driver.
        :param node_: Node that is being executed
        :param kwargs: Keyword arguments that are being passed into the node
        :param task_id: ID of the task, defaults to None if not in a task setting
        """
        pass


@lifecycle.base_method("do_node_execute")
class BaseDoNodeExecuteAsync(abc.ABC):
    @abc.abstractmethod
    async def do_node_execute(
        self,
        *,
        run_id: str,
        node_: "node.Node",
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
    ) -> Any:
        """Asynchronous method that is called to implement node execution. This can replace the execution of a node
        with something all together, augment it, or delegate it.

        :param run_id: ID of the run, unique in scope of the driver.
        :param node_: Node that is being executed
        :param kwargs: Keyword arguments that are being passed into the node
        :param task_id: ID of the task, defaults to None if not in a task setting
        """
        pass


@lifecycle.base_hook("post_node_execute")
class BasePostNodeExecute(abc.ABC):
    @abc.abstractmethod
    def post_node_execute(
        self,
        *,
        run_id: str,
        node_: "node.Node",
        kwargs: Dict[str, Any],
        success: bool,
        error: Optional[Exception],
        result: Optional[Any],
        task_id: Optional[str] = None,
    ):
        """Hook that is called immediately after node execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param node_: Node that is being executed
        :param kwargs: Keyword arguments that are being passed into the node
        :param success: Whether or not the node executed successfully
        :param error: The error that was raised, if any
        :param result: The result of the node execution, if no error was raised
        :param task_id: ID of the task, defaults to None if not in a task-based execution
        """
        pass


@lifecycle.base_hook("post_node_execute")
class BasePostNodeExecuteAsync(abc.ABC):
    @abc.abstractmethod
    async def post_node_execute(
        self,
        *,
        run_id: str,
        node_: "node.Node",
        kwargs: Dict[str, Any],
        success: bool,
        error: Optional[Exception],
        result: Any,
        task_id: Optional[str] = None,
    ):
        """Hook that is called immediately after node execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param node_: Node that is being executed
        :param kwargs: Keyword arguments that are being passed into the node
        :param success: Whether or not the node executed successfully
        :param error: The error that was raised, if any
        :param result: The result of the node execution, if no error was raised
        :param task_id: ID of the task, defaults to None if not in a task-based execution
        """
        pass


@lifecycle.base_hook("post_task_execute")
class BasePostTaskExecute(abc.ABC):
    @abc.abstractmethod
    def post_task_execute(
        self,
        *,
        run_id: str,
        task_id: str,
        nodes: List["node.Node"],
        results: Optional[Dict[str, Any]],
        success: bool,
        error: Exception,
    ):
        """Hook called immediately after task execution. Note that this is only useful in dynamic
        execution, although we reserve the right to add this back into the standard hamilton execution pattern.

        :param run_id: ID of the run, unique in scope of the driver.
        :param task_id: ID of the task
        :param nodes: Nodes that were executed
        :param results: Results of the task
        :param success: Whether or not the task executed successfully
        :param error: The error that was raised, if any
        """
        pass


@lifecycle.base_hook("post_task_execute")
class BasePostTaskExecuteAsync(abc.ABC):
    @abc.abstractmethod
    async def post_task_execute(
        self,
        *,
        run_id: str,
        task_id: str,
        nodes: List["node.Node"],
        results: Optional[Dict[str, Any]],
        success: bool,
        error: Exception,
    ):
        """Asynchronous Hook called immediately after task execution. Note that this is only useful in dynamic
        execution, although we reserve the right to add this back into the standard hamilton execution pattern.

        :param run_id: ID of the run, unique in scope of the driver.
        :param task_id: ID of the task
        :param nodes: Nodes that were executed
        :param results: Results of the task
        :param success: Whether or not the task executed successfully
        :param error: The error that was raised, if any
        """
        pass


@lifecycle.base_hook("post_graph_execute")
class BasePostGraphExecute(abc.ABC):
    @abc.abstractmethod
    def post_graph_execute(
        self,
        *,
        run_id: str,
        graph: "graph.FunctionGraph",
        success: bool,
        error: Optional[Exception],
        results: Optional[Dict[str, Any]],
    ):
        """Hook called immediately after graph execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param graph: Graph that was executed
        :param success: Whether or not the graph executed successfully
        :param error: Error that was raised, if any
        :param results: Results of the graph execution
        """
        pass


@lifecycle.base_hook("post_graph_execute")
class BasePostGraphExecuteAsync(abc.ABC):
    @abc.abstractmethod
    async def post_graph_execute(
        self,
        *,
        run_id: str,
        graph: "graph.FunctionGraph",
        success: bool,
        error: Optional[Exception],
        results: Optional[Dict[str, Any]],
    ):
        """Asynchronous Hook called immediately after graph execution.

        :param run_id: ID of the run, unique in scope of the driver.
        :param graph: Graph that was executed
        :param success: Whether or not the graph executed successfully
        :param error: Error that was raised, if any
        :param results: Results of the graph execution
        """
        pass


@lifecycle.base_method("do_build_result")
class BaseDoBuildResult(abc.ABC):
    @abc.abstractmethod
    def do_build_result(self, *, outputs: Any) -> Any:
        """Method that is called to build the result of the graph execution.

        :param outputs: Output of the node execution
        :return: The final result
        """
        pass


# This is the type of a lifecycle adapter -- these types utilize

LifecycleAdapter = Union[
    BasePreDoAnythingHook,
    BaseDoCheckEdgeTypesMatch,
    BaseDoValidateInput,
    BaseValidateNode,
    BaseValidateGraph,
    BasePostGraphConstruct,
    BasePostGraphConstructAsync,
    BasePreGraphExecute,
    BasePreGraphExecuteAsync,
    BasePreTaskExecute,
    BasePreTaskExecuteAsync,
    BasePreNodeExecute,
    BasePreNodeExecuteAsync,
    BaseDoNodeExecute,
    BaseDoNodeExecuteAsync,
    BasePostNodeExecute,
    BasePostNodeExecuteAsync,
    BasePostTaskExecute,
    BasePostTaskExecuteAsync,
    BasePostGraphExecute,
    BasePostGraphExecuteAsync,
    BaseDoBuildResult,
]


class LifecycleAdapterSet:
    """An internal class that groups together all the lifecycle adapters.
    This allows us to call methods through a delegation pattern, enabling us to add
    whatever callbacks, logging, error-handling, etc... we need globally. While this
    does increase the stack trace in an error, it should be pretty easy to figure out what'g going on.
    """

    def __init__(self, *adapters: LifecycleAdapter):
        """Initializes the adapter set.

        :param adapters: Adapters to group together
        """
        self._adapters = self._uniqify_adapters(adapters)
        self.sync_hooks, self.async_hooks = self._get_lifecycle_hooks()
        self.sync_methods, self.async_methods = self._get_lifecycle_methods()
        self.sync_validators = self._get_lifecycle_validators()

    def _uniqify_adapters(self, adapters: List[LifecycleAdapter]) -> List[LifecycleAdapter]:
        """Removes duplicate adapters from the list of adapters -- this often happens on how they're passed in
        and we don't want to have the same adapter twice. Specifically, this came up due to parsing/splitting out adapters
        with async lifecycle hooks -- there were cases in which we were passed duplicates. This was compounded as we would pass
        adapters to other adapter sets and end up further duplicating.

        TODO -- remove this and ensure that no case passes in duplicates.
        """

        seen = set()
        return [
            adapter for adapter in adapters if not (id(adapter) in seen or seen.add(id(adapter)))
        ]

    def _get_lifecycle_validators(
        self,
    ) -> Dict[str, List[LifecycleAdapter]]:
        sync_validators = collections.defaultdict(set)
        for adapter in self.adapters:
            for cls in inspect.getmro(adapter.__class__):
                sync_validator = getattr(cls, SYNC_VALIDATOR, None)
                if sync_validator is not None:
                    sync_validators[sync_validator].add(adapter)
        return {validator: list(adapters) for validator, adapters in sync_validators.items()}

    def _get_lifecycle_hooks(
        self,
    ) -> Tuple[Dict[str, List[LifecycleAdapter]], Dict[str, List[LifecycleAdapter]]]:
        sync_hooks = collections.defaultdict(list)
        async_hooks = collections.defaultdict(list)
        for adapter in self.adapters:
            for cls in inspect.getmro(adapter.__class__):
                sync_hook = getattr(cls, SYNC_HOOK, None)
                if sync_hook is not None:
                    if adapter not in sync_hooks[sync_hook]:
                        sync_hooks[sync_hook].append(adapter)
                async_hook = getattr(cls, ASYNC_HOOK, None)
                if async_hook is not None:
                    if adapter not in async_hooks[async_hook]:
                        async_hooks[async_hook].append(adapter)
        return (
            {hook: adapters for hook, adapters in sync_hooks.items()},
            {hook: adapters for hook, adapters in async_hooks.items()},
        )

    def _get_lifecycle_methods(
        self,
    ) -> Tuple[Dict[str, List[LifecycleAdapter]], Dict[str, List[LifecycleAdapter]]]:
        sync_methods = collections.defaultdict(set)
        async_methods = collections.defaultdict(set)
        for adapter in self.adapters:
            for cls in inspect.getmro(adapter.__class__):
                sync_method = getattr(cls, SYNC_METHOD, None)
                if sync_method is not None:
                    sync_methods[sync_method].add(adapter)
                async_method = getattr(cls, ASYNC_METHOD, None)
                if async_method is not None:
                    async_methods[async_method].add(adapter)
        multiple_implementations_sync = [
            method for method, adapters in sync_methods.items() if len(adapters) > 1
        ]
        multiple_implementations_async = [
            method for method, adapters in async_methods.items() if len(adapters) > 1
        ]
        if len(multiple_implementations_sync) > 0 or len(multiple_implementations_async) > 0:
            raise ValueError(
                f"Multiple adapters cannot (currently) implement the same lifecycle method. "
                f"Sync methods: {multiple_implementations_sync}. "
                f"Async methods: {multiple_implementations_async}"
            )
        return (
            {method: list(adapters) for method, adapters in sync_methods.items()},
            {method: list(adapters) for method, adapters in async_methods.items()},
        )

    def does_hook(self, hook_name: str, is_async: Optional[bool] = None) -> bool:
        """Whether or not a hook is implemented by any of the adapters in this group.
        If this hook is not registered, this will raise a ValueError.

        :param hook_name: Name of the hook
        :param is_async: Whether you want the async version or not
        :return: True if this adapter set does this hook, False otherwise
        """
        either = is_async is None
        if (is_async or either) and hook_name not in REGISTERED_ASYNC_HOOKS:
            raise ValueError(
                f"Hook {hook_name} is not registered as an asynchronous lifecycle hook. "
                f"Registered hooks are {REGISTERED_ASYNC_HOOKS}"
            )
        if ((not is_async) or either) and hook_name not in REGISTERED_SYNC_HOOKS:
            raise ValueError(
                f"Hook {hook_name} is not registered as a synchronous lifecycle hook. "
                f"Registered hooks are {REGISTERED_SYNC_HOOKS}"
            )
        has_async = hook_name in self.async_hooks
        has_sync = hook_name in self.sync_hooks
        return (has_async or has_sync) if either else has_async if is_async else has_sync

    def does_method(self, method_name: str, is_async: Optional[bool] = None) -> bool:
        """Whether a method is implemented by any of the adapters in this group.
        If this method is not registered, this will raise a ValueError.

        :param method_name: Name of the method
        :param is_async: Whether you want the async version or not
        :return: True if this adapter set does this method, False otherwise
        """
        either = is_async is None
        if (is_async or either) and method_name not in REGISTERED_ASYNC_METHODS:
            raise ValueError(
                f"Method {method_name} is not registered as an asynchronous lifecycle method. "
                f"Registered methods are {REGISTERED_ASYNC_METHODS}"
            )
        if ((not is_async) or either) and method_name not in REGISTERED_SYNC_METHODS:
            raise ValueError(
                f"Method {method_name} is not registered as a synchronous lifecycle method. "
                f"Registered methods are {REGISTERED_SYNC_METHODS}"
            )
        has_async = method_name in self.async_methods
        has_sync = method_name in self.sync_methods
        return (has_async or has_sync) if either else has_async if is_async else has_sync

    def does_validation(self, validator_name: str) -> bool:
        """Whether a validator is implemented by any of the adapters in this group.

        :param validator_name: Name of the validator
        :param is_async: Whether you want the async version or not
        :return: True if this adapter set does this validator, False otherwise
        """
        if validator_name not in REGISTERED_SYNC_VALIDATORS:
            raise ValueError(
                f"Validator {validator_name} is not registered as a lifecycle validator. "
                f"Registered validators are {REGISTERED_SYNC_VALIDATORS}"
            )
        return validator_name in self.sync_validators

    def call_all_lifecycle_hooks_sync(self, hook_name: str, **kwargs):
        """Calls all the lifecycle hooks in this group, by hook name (stage)

        :param hook_name: Name of the hooks to call
        :param kwargs: Keyword arguments to pass into the hook
        """
        for adapter in self.sync_hooks.get(hook_name, []):
            getattr(adapter, hook_name)(**kwargs)

    async def call_all_lifecycle_hooks_async(self, hook_name: str, **kwargs):
        """Calls all the lifecycle hooks in this group, by hook name (stage).

        :param hook_name: Name of the hook
        :param kwargs: Keyword arguments to pass into the hook
        """
        futures = []
        for adapter in self.async_hooks.get(hook_name, []):
            futures.append(getattr(adapter, hook_name)(**kwargs))
        await asyncio.gather(*futures)

    async def call_all_lifecycle_hooks_sync_and_async(self, hook_name: str, **kwargs):
        """Calls all the lifecycle hooks whether they are sync or async

        :param hook_name: name of the hook
        :param kwargs: keyword arguments for the hook
        """
        self.call_all_lifecycle_hooks_sync(hook_name, **kwargs)
        await self.call_all_lifecycle_hooks_async(hook_name, **kwargs)

    def call_lifecycle_method_sync(self, method_name: str, **kwargs) -> Any:
        """Calls a lifecycle method in this group, by method name.

        :param method_name: Name of the method
        :param kwargs: Keyword arguments to pass into the method
        :return: The result of the method
        """
        if method_name not in REGISTERED_SYNC_METHODS:
            raise ValueError(
                f"Method {method_name} is not registered as a synchronous lifecycle method. "
                f"Registered methods are {REGISTERED_SYNC_METHODS}"
            )
        if method_name not in self.sync_methods:
            raise ValueError(
                f"Method {method_name} is not implemented by any of the adapters in this group. "  # TODO _- improve the error message
                f"Registered methods are {self.sync_methods}"
            )
        (adapter,) = self.sync_methods[method_name]
        return getattr(adapter, method_name)(**kwargs)

    async def call_lifecycle_method_async(self, method_name: str, **kwargs):
        """Call a lifecycle method in this group, by method name, async

        :param method_name: Name of the method
        :param kwargs: Keyword arguments to pass into the method
        :return: The result of the method
        """
        if method_name not in REGISTERED_ASYNC_METHODS:
            raise ValueError(
                f"Method {method_name} is not registered as an asynchronous lifecycle method. "
                f"Registered methods are {REGISTERED_ASYNC_METHODS}"
            )
        (adapter,) = self.async_methods[method_name]
        return await getattr(adapter, method_name)(**kwargs)

    def call_all_validators_sync(
        self, validator_name: str, output_only_failures: bool = True, **kwargs
    ) -> List[ValidationResult]:
        """Calls all the lifecycle validators in this group, by validator name (stage)

        :param validator_name: Name of the validators to call
        :param kwargs: Keyword arguments to pass into the validator
        :param output_only_failures: Whether to output only failures
        """
        results = []
        for adapter in self.sync_validators[validator_name]:
            is_valid, message = getattr(adapter, validator_name)(**kwargs)
            if not is_valid or not output_only_failures:
                results.append(ValidationResult(success=is_valid, error=message, validator=adapter))
        return results

    @property
    def adapters(self) -> List[LifecycleAdapter]:
        """Gives the adapters in this group

        :return: A list of adapters
        """
        return self._adapters

    async def ainit(self):
        """Asynchronously initializes the adapters in this group. This is so we can avoid having an async constructor
        -- it is an implicit internal-facing contract -- the async adapters are allowed one ainit()
        method that will be called by the driver.

        Note this is not public-facing -- E.G. you cannot expect to define this on your own adapters. We may consider adding
        a ``pre_do_anything`` async hook and removing this, but for now this should suffice.
        """
        for adapter in self.adapters:
            if hasattr(adapter, "ainit"):
                await adapter.ainit()



import concurrent
import logging
import tempfile
from typing import Generator, Union

import dp

logger = logging.getLogger(__name__)

from hamilton import contrib

with contrib.catch_import_errors(__name__, __file__, logger):
    import openai
    import tiktoken
    from pypdf import PdfReader
    from tenacity import retry, stop_after_attempt, wait_random_exponential
    from tqdm import tqdm

from hamilton.function_modifiers import config


def summarize_chunk_of_text_prompt(content_type: str = "an academic paper") -> str:
    """Base prompt for summarizing chunks of text."""
    return f"Summarize this text from {content_type}. Extract any key points with reasoning.\n\nContent:"


def summarize_text_from_summaries_prompt(
    content_type: str = "an academic paper",
) -> str:
    """Prompt for summarizing a paper from a list of summaries."""
    return f"""Write a summary collated from this collection of key points extracted from {content_type}.
    The summary should highlight the core argument, conclusions and evidence, and answer the user's query.
    User query: {{query}}
    The summary should be structured in bulleted lists following the headings Core Argument, Evidence, and Conclusions.
    Key points:\n{{results}}\nSummary:\n"""


@config.when(file_type="pdf")
def raw_text__pdf(pdf_source: Union[str, bytes, tempfile.SpooledTemporaryFile]) -> str:
    """Takes a filepath to a PDF and returns a string of the PDF's contents
    :param pdf_source: the path, or the temporary file, to the PDF.
    :return: the text of the PDF.
    """
    reader = PdfReader(pdf_source)
    _pdf_text = ""
    page_number = 0
    for page in reader.pages:
        page_number += 1
        _pdf_text += page.extract_text() + f"\nPage Number: {page_number}"
    return _pdf_text


@config.when(file_type="txt")
def raw_text__txt(text_file: Union[str, tempfile.SpooledTemporaryFile]) -> str:
    """Takes a filepath to a text file and returns a string of the text file's contents
    :param text_file: the path, or the temporary file, to the text file.
    :return: the contents of the file as a string.
    """
    with open(text_file) as f:
        return f.read()


def _create_chunks(text: str, n: int, tokenizer: tiktoken.Encoding) -> Generator[str, None, None]:
    """Helper function. Returns successive n-sized chunks from provided text.
    Split a text into smaller chunks of size n, preferably ending at the end of a sentence
    :param text:
    :param n:
    :param tokenizer:
    :return:
    """
    tokens = tokenizer.encode(text)
    i = 0
    while i < len(tokens):
        # Find the nearest end of sentence within a range of 0.5 * n and 1.5 * n tokens
        j = min(i + int(1.5 * n), len(tokens))
        while j > i + int(0.5 * n):
            # Decode the tokens and check for full stop or newline
            chunk = tokenizer.decode(tokens[i:j])
            if chunk.endswith(".") or chunk.endswith("\n"):
                break
            j -= 1
        # If no end of sentence found, use n tokens as the chunk size
        if j == i + int(0.5 * n):
            j = min(i + n, len(tokens))
        yield tokens[i:j]
        i = j


def chunked_text(
    raw_text: str, tokenizer_encoding: str = "cl100k_base", max_token_length: int = 1500
) -> list[str]:
    """Chunks the pdf text into smaller chunks of size max_token_length.
    :param raw_text: the Series of individual pdf texts to chunk.
    :param max_token_length: the maximum length of tokens in each chunk.
    :param tokenizer_encoding: the encoding to use for the tokenizer.
    :return: Series of chunked pdf text. Each element is a list of chunks.
    """
    tokenizer = tiktoken.get_encoding(tokenizer_encoding)
    _encoded_chunks = _create_chunks(raw_text, max_token_length, tokenizer)
    _decoded_chunks = [tokenizer.decode(chunk) for chunk in _encoded_chunks]
    return _decoded_chunks


@retry(wait=wait_random_exponential(min=1, max=40), stop=stop_after_attempt(3))
def _summarize_chunk(content: str, template_prompt: str, openai_gpt_model: str) -> str:
    """This function applies a prompt to some input content. In this case it returns a summarized chunk of text.
    :param content: the content to summarize.
    :param template_prompt: the prompt template to use to put the content into.
    :param openai_gpt_model: the openai gpt model to use.
    :return: the response from the openai API.
    """
    prompt = template_prompt + content
    response = openai.ChatCompletion.create(
        model=openai_gpt_model,
        messages=[{"role": "user", "content": prompt}],
        temperature=0,
    )
    return response["choices"][0]["message"]["content"]


def summarized_chunks(
    chunked_text: list[str], summarize_chunk_of_text_prompt: str, openai_gpt_model: str
) -> str:
    """Summarizes a series of chunks of text.
    Note: this takes the first result from the top_n_related_articles series and summarizes it. This is because
    the top_n_related_articles series is sorted by relatedness, so the first result is the most related.
    :param chunked_text: a list of chunks of text for an article.
    :param summarize_chunk_of_text_prompt:  the prompt to use to summarize each chunk of text.
    :param openai_gpt_model: the openai gpt model to use.
    :return: a single string of each chunk of text summarized, concatenated together.
    """
    _summarized_text = ""
    with concurrent.futures.ThreadPoolExecutor(max_workers=len(chunked_text)) as executor:
        futures = [
            executor.submit(
                _summarize_chunk,
                chunk,
                summarize_chunk_of_text_prompt,
                openai_gpt_model,
            )
            for chunk in chunked_text
        ]
        with tqdm(total=len(chunked_text)) as pbar:
            for _ in concurrent.futures.as_completed(futures):
                pbar.update(1)
        for future in futures:
            data = future.result()
            _summarized_text += data
    return _summarized_text


def prompt_and_text_content(
    summarized_chunks: str,
    summarize_text_from_summaries_prompt: str,
    user_query: str,
) -> str:
    """Creates the prompt for summarizing the text from the summarized chunks of the pdf.
    :param summarized_chunks: a long string of chunked summaries of a file.
    :param summarize_text_from_summaries_prompt: the template to use to summarize the chunks.
    :param user_query: the original user query.
    :return: the prompt to use to summarize the chunks.
    """
    return summarize_text_from_summaries_prompt.format(query=user_query, results=summarized_chunks)


def summarized_text(
    prompt_and_text_content: str,
    openai_gpt_model: str,
) -> str:
    """Summarizes the text from the summarized chunks of the pdf.
    :param prompt_and_text_content: the prompt and content to send over.
    :param openai_gpt_model: which openai gpt model to use.
    :return: the string response from the openai API.
    """
    response = openai.ChatCompletion.create(
        model=openai_gpt_model,
        messages=[
            {
                "role": "user",
                "content": prompt_and_text_content,
            }
        ],
        temperature=0,
    )
    return response["choices"][0]["message"]["content"]


if __name__ == "__main__":
    # run as a script to test Hamilton's execution
    import __init__ as text_summarization

    from hamilton import base, driver

    dr = driver.Driver(
        {"file_type": "pdf"},
        text_summarization,
        adapter=base.DefaultAdapter(),
    )
    # create the DAG image
    dr.display_all_functions("dag", {"format": "png", "view": False}, orient="TB")



import inspect
import json
import sys
import textwrap

import pytest

from hamilton import driver, graph_types, node
from hamilton.node import Node, NodeType

from tests import nodes as test_nodes
from tests.resources.dynamic_parallelism import no_parallel


@pytest.fixture()
def func_a():
    """Default function implementation"""

    return textwrap.dedent(
        """
        def A(external_input: int) -> int:
            return external_input % 7
        """
    )


@pytest.fixture()
def func_a_body():
    """The function A() has modulo 5 instead of 7"""

    return textwrap.dedent(
        """
        def A(external_input: int) -> int:
            return external_input % 5
        """
    )


@pytest.fixture()
def func_a_docstring():
    """The function A() has a docstring"""

    return textwrap.dedent(
        """
        def A(external_input: int) -> int:
            '''This one has a docstring'''
            return external_input % 7
        """
    )


@pytest.fixture()
def func_a_multiline():
    """The function A() has a docstring"""

    return textwrap.dedent(
        """
        def A(external_input: int) -> int:
            '''This one has
            a multiline
            docstring
            '''
            return external_input % 7
        """
    )


@pytest.fixture()
def func_a_comment():
    """The function A() has a comment"""

    return textwrap.dedent(
        """
        def A(external_input: int) -> int:
            return external_input % 7  # a comment
        """
    )


def test_create_hamilton_node():
    def node_to_create(required_dep: int, optional_dep: int = 1) -> str:
        """Documentation"""
        return f"{required_dep}_{optional_dep}"

    n = node.Node.from_fn(
        node_to_create
    ).copy_with(  # The following simulate the graph's creation of a node
        tags={"tag_key": "tag_value"}, originating_functions=(node_to_create,)
    )
    hamilton_node = graph_types.HamiltonNode.from_node(n)
    assert hamilton_node.name == "node_to_create"
    assert hamilton_node.type == str
    assert hamilton_node.tags["tag_key"] == "tag_value"
    assert hamilton_node.originating_functions == (node_to_create,)
    assert hamilton_node.documentation == "Documentation"
    assert not hamilton_node.is_external_input
    assert hamilton_node.required_dependencies == {"required_dep"}
    assert hamilton_node.optional_dependencies == {"optional_dep"}
    assert hamilton_node.version == graph_types.hash_source_code(node_to_create, strip=True)

    assert hamilton_node.as_dict() == {
        "name": "node_to_create",
        "tags": {"tag_key": "tag_value"},
        "output_type": "str",
        "required_dependencies": ["required_dep"],
        "optional_dependencies": ["optional_dep"],
        "source": (
            "    def node_to_create(required_dep: int, optional_dep: int = 1) -> str:\n"
            '        """Documentation"""\n'
            '        return f"{required_dep}_{optional_dep}"\n'
        ),
        "documentation": "Documentation",
        "version": graph_types.hash_source_code(node_to_create, strip=True),
    }


def test_create_hamilton_config_node_version():
    """Config nodes now return the name as the version."""
    n = Node("foo", int, node_source=NodeType.EXTERNAL)
    hamilton_node = graph_types.HamiltonNode.from_node(n)
    # the above will have no specified versions
    assert hamilton_node.version == "foo"
    assert hamilton_node.as_dict()["version"] == "foo"


def test_create_hamilton_node_missing_version():
    """We contrive a case where originating_functions is None."""

    def foo(i: int) -> int:
        return i

    n = Node("foo", int, "", foo, node_source=NodeType.STANDARD)
    hamilton_node = graph_types.HamiltonNode.from_node(n)
    # the above will have no specified versions
    assert hamilton_node.version is None
    assert hamilton_node.as_dict()["version"] is None


def test_hamilton_graph_version_normal():
    dr = driver.Builder().with_modules(no_parallel).build()
    graph = graph_types.HamiltonGraph.from_graph(dr.graph)
    # assumption is for python 3
    if sys.version_info.minor == 8:
        hash_value = "0a375f3366590453dea8927d4c02c15dc090f8be42e9129d9a1139284eac920c"
    else:
        hash_value = "3b3487599ccc4fc56995989c6d32b58a90c0b91b8c16b3f453a2793f47436831"
    assert graph.version == hash_value


def test_hamilton_graph_version_with_none_originating_functions():
    dr = driver.Builder().with_modules(no_parallel).build()
    graph = graph_types.HamiltonGraph.from_graph(dr.graph)
    # if this gets flakey we should find a specific node to make None
    graph.nodes[-1].originating_functions = None
    # assumption is for python 3
    if sys.version_info.minor == 8:
        hash_value = "7d556424cd84b97a395d9b6219f502e8f818f17002ce88f47974266c0cce454a"
    else:
        hash_value = "781d89517c1744c40a7afcdc49ee8592fbb23955e28d87f1b584d08430a3e837"
    assert graph.version == hash_value


def test_hamilton_graph_getitem():
    dr = driver.Builder().with_modules(test_nodes).build()
    h_graph = graph_types.HamiltonGraph.from_graph(dr.graph)

    assert isinstance(h_graph["lowercased"], graph_types.HamiltonNode)
    assert h_graph["lowercased"].name == "lowercased"


def test_hamilton_graph_filter_nodes():
    dr = driver.Builder().with_modules(test_nodes).build()
    h_graph = graph_types.HamiltonGraph.from_graph(dr.graph)

    # could be a lambda but flake8 doesn't like it.
    def filter_cache_equal_str(n: graph_types.HamiltonNode) -> bool:
        return n.tags.get("cache") == "str"

    node_selection = h_graph.filter_nodes(filter=filter_cache_equal_str)
    node_names = [n.name for n in node_selection]

    assert len(node_selection) == 2
    assert "lowercased" in node_names
    assert "uppercased" in node_names


def test_json_serializable_dict():
    for name, obj in inspect.getmembers(test_nodes):
        if inspect.isfunction(obj) and not name.startswith("_"):
            n = node.Node.from_fn(
                obj
            ).copy_with(  # The following simulate the graph's creation of a node
                originating_functions=(obj,)
            )
            hamilton_node = graph_types.HamiltonNode.from_node(n)

            # Check that json.dumps works on all nodes
            json.dumps(hamilton_node.as_dict())


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires Python 3.9 or higher")
def test_remove_docstring(func_a: str, func_a_docstring: str):
    func_a_no_whitespace = func_a.strip()
    stripped = graph_types._remove_docs_and_comments(func_a_docstring)
    assert func_a_no_whitespace == stripped


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires Python 3.9 or higher")
def test_remove_multiline(func_a: str, func_a_multiline: str):
    func_a_no_whitespace = func_a.strip()
    stripped = graph_types._remove_docs_and_comments(func_a_multiline)
    assert func_a_no_whitespace == stripped


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires Python 3.9 or higher")
def test_remove_comment(func_a: str, func_a_comment: str):
    func_a_no_whitespace = func_a.strip()
    stripped = graph_types._remove_docs_and_comments(func_a_comment)
    assert func_a_no_whitespace == stripped


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires Python 3.9 or higher")
@pytest.mark.parametrize("strip", [True, False])
def test_different_hash_function_body(func_a: str, func_a_body: str, strip: bool):
    """Gives different hash for different function body"""
    func_a_hash = graph_types.hash_source_code(func_a, strip=strip)
    func_a_body_hash = graph_types.hash_source_code(func_a_body, strip=strip)
    assert func_a_hash != func_a_body_hash


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires Python 3.9 or higher")
@pytest.mark.parametrize("strip", [True, False])
def test_different_hash_docstring(func_a: str, func_a_docstring: str, strip: bool):
    """Same hash if strip docstring, else different hash"""
    func_a_hash = graph_types.hash_source_code(func_a, strip=strip)
    func_a_docstring_hash = graph_types.hash_source_code(func_a_docstring, strip=strip)
    assert (func_a_hash == func_a_docstring_hash) is (True if strip else False)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires Python 3.9 or higher")
@pytest.mark.parametrize("strip", [True, False])
def test_different_hash_multiline_docstring(func_a: str, func_a_multiline: str, strip: bool):
    """Same hash if strip multiline docstring, else different hash"""
    func_a_hash = graph_types.hash_source_code(func_a, strip=strip)
    func_a_multiline_hash = graph_types.hash_source_code(func_a_multiline, strip=strip)
    assert (func_a_hash == func_a_multiline_hash) is (True if strip else False)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires Python 3.9 or higher")
@pytest.mark.parametrize("strip", [True, False])
def test_different_hash_comment(func_a: str, func_a_comment: str, strip: bool):
    """Same hash if strip comment, else different hash"""
    func_a_hash = graph_types.hash_source_code(func_a, strip=strip)
    func_a_comment = graph_types.hash_source_code(func_a_comment, strip=strip)
    assert (func_a_hash == func_a_comment) is (True if strip else False)



import pandas as pd
from evidently.metric_preset import DataDriftPreset
from evidently.report import Report
from evidently.test_preset import DataStabilityTestPreset
from evidently.test_suite import TestSuite
from sklearn import datasets

iris_data = datasets.load_iris(as_frame="auto")
iris_frame = iris_data.frame

data_stability = TestSuite(
    tests=[
        DataStabilityTestPreset(),
    ]
)
data_stability.run(
    current_data=iris_frame.iloc[:60],
    reference_data=iris_frame.iloc[60:],
    column_mapping=None,
)
data_stability



import argparse
import ast
import json
import os
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Set, Tuple, Union

from IPython.core.magic import Magics, cell_magic, line_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from IPython.core.shellapp import InteractiveShellApp
from IPython.display import HTML, Code, display
from IPython.utils.process import arg_split

from hamilton import ad_hoc_utils, driver


def get_assigned_variables(module_node: ast.Module) -> Set[str]:
    """Get the set of variable names assigned in a AST Module"""
    assigned_vars = set()

    def visit_node(ast_node):
        """Recursive function looking for assigned variable names"""
        if isinstance(ast_node, ast.Assign):
            for target in ast_node.targets:
                if isinstance(target, ast.Name):
                    assigned_vars.add(target.id)

        for child_node in ast.iter_child_nodes(ast_node):
            visit_node(child_node)

    visit_node(module_node)
    return assigned_vars


def execute_and_get_assigned_values(shell: InteractiveShellApp, cell: str) -> Dict[str, Any]:
    """Execute source code from a cell in the user namespace and collect
    the values of all assigned variables into a dictionary.
    """
    shell.ex(cell)
    expr = shell.input_transformer_manager.transform_cell(cell)
    expr_ast = shell.compile.ast_parse(expr)
    return {name: shell.user_ns[name] for name in get_assigned_variables(expr_ast)}


def topological_sort(nodes):
    """Sort the nodes for nice output display"""

    def dfs(node, visited, stack):
        visited.add(node)
        for neighbor in graph.get(node, []):
            if neighbor not in visited:
                dfs(neighbor, visited, stack)
        stack.append(node)

    graph = {n.name: set([*n.required_dependencies, *n.optional_dependencies]) for n in nodes}
    visited = set()
    stack = []

    for node in graph:
        if node not in visited:
            dfs(node, visited, stack)

    return stack


def _normalize_result_names(node_name: str) -> str:
    """Remove periods from the name of dynamically generated Hamilton nodes"""
    return node_name.replace(".", "__")


def display_in_databricks(dot):
    try:
        display(HTML(dot.pipe(format="svg").decode("utf-8")))
    except Exception as e:
        print(
            f"Failed to display graph: {e}\n"
            "Please ensure graphviz is installed via `%sh apt install -y graphviz`"
        )
        return
    return dot


def insert_cell_with_content():
    """Calling this function before .set_next_input() will output text content to
    the next code cell.

    This works by printing JavaScript in the current cell's HTML output

    adapted from: https://stackoverflow.com/questions/65379879/define-a-ipython-magic-which-replaces-the-content-of-the-next-cell
    """
    js_script = r"""<script>

        if (document.getElementById('notebook-container')) {
            //console.log('Jupyter Notebook');
            allCells = document.getElementById('notebook-container').children;
            selectionClass = /\bselected\b/;
            jupyter = 'notebook';
        }
        else if (document.getElementsByClassName('jp-Notebook-cell').length){
            //console.log('Jupyter Lab');
            allCells = document.getElementsByClassName('jp-Notebook-cell');
            selectionClass = /\bjp-mod-selected\b/;
            jupyter = 'lab';
        }
        else {
            console.log('Unknown Environment');
        }

        if (typeof allCells !== 'undefined') {
            for (i = 0; i < allCells.length - 1; i++) {
                if(selectionClass.test(allCells[i].getAttribute('class'))){
                    allCells[i + 1].remove();

                    // remove output indicators of current cell
                    window.setTimeout(function(){
                        if(jupyter === 'lab') {
                            allCells[i].setAttribute('class', allCells[i].getAttribute('class') + ' jp-mod-noOutputs');
                            allCells[i].getElementsByClassName('jp-OutputArea jp-Cell-outputArea')[0].innerHTML = '';
                        } else if(jupyter === 'notebook'){
                            allCells[i].getElementsByClassName('output')[0].innerHTML = '';
                        }
                    }, 20);

                    break;
                }
            }
        }
        </script>"""

    # remove next cell
    display(HTML(js_script))


def determine_notebook_type() -> str:
    if "DATABRICKS_RUNTIME_VERSION" in os.environ:
        return "databricks"
    return "default"


def parse_known_argstring(magic_func, argstring) -> Tuple[argparse.Namespace, List[str]]:
    """IPython magic arguments parsing doesn't allow unknown args.
    Used instead of IPython.core.magic_arguments.parse_argstring

    IPython ref: https://github.com/ipython/ipython/blob/43781b39a67f02ff4e9ae63484387f654dd045d4/IPython/core/magic_arguments.py#L164
    argparse ref: https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.parse_known_args
    """
    argv = arg_split(argstring)
    # magic_func.parser is an argparse.ArgumentParser subclass
    known, unknown = magic_func.parser.parse_known_args(argv)
    return known, unknown


def parse_key_value_config(config_string):
    config = {}
    for item in config_string.split():
        key, value = item.split("=")
        config[key] = value
    return config


@magics_class
class HamiltonMagics(Magics):
    """Magics to facilitate interactive Hamilton development in notebooks."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.builder = None
        self.notebook_env = determine_notebook_type()
        self.incremental_cells_state = defaultdict(dict)

    def resolve_unknown_args_cell_to_module(self, unknown: List[str]):
        """Handle unknown arguments. It won't make the magic execution fail."""

        # deprecated in V2 because it's less useful since `%%cell_to_module` can execute itself
        if any(arg in ("-r", "--rebuilder-drivers") for arg in unknown):
            print(
                "DeprecationWarning: -r/--rebuilder-drivers no long does anything and will be removed in future releases."
            )

        # deprecated in V2 because it relates to the deprecated -r/--rebuild-drivers
        if any(arg in ("-v", "--verbose") for arg in unknown):
            print(
                "DeprecationWarning: -v/--verbose no long does anything and will be removed in future releases."
            )

        # there for backwards compatibility. Equivalent to calling `%%cell_to_module?`
        # not included as @argument because it's not really a function arg to %%cell_to_module
        if any(arg in ("-h", "--help") for arg in unknown):
            print(help(self.cell_to_module))

    def resolve_config_arg(self, config_arg) -> Union[bool, dict]:
        # default case: didn't receive `-c/--config`. Set an empty dict
        if config_arg is None:
            config = {}
        # case 1, 2, 3: `-c/--config` is specified
        # case 1: -c/--config refers to variable in the user namespace
        elif self.shell.user_ns.get(config_arg):
            config = self.shell.user_ns.get(config_arg)
        # case 2: parse using key=value
        elif "=" in config_arg:
            config = parse_key_value_config(config_arg)
        # case 3: parse as JSON
        elif ":" in config_arg:
            try:
                # strip quotation marks added by IPython and avoid mutating `args`
                config_str = config_arg.strip("'\"")
                config = json.loads(config_str)
            except json.JSONDecodeError:
                print(f"JSONDecodeError: Failed to parse `config` as JSON. Received {config_arg}")
                return False
        return config

    @magic_arguments()  # needed on top to enable parsing
    @argument("name", nargs="?", help="Name for the module defined in this cell.")
    @argument(
        "-m",
        "--module_name",
        nargs="?",
        const=True,
        help="Alias for positional argument `module_name`. There for backwards compatibility. Prefer the position arg.",
    )
    @argument(
        "-d",
        "--display",
        nargs="?",
        const=True,
        help="Display the dataflow. The argument is the variable name of a dictionary of visualization kwargs; else {}.",
    )
    @argument(
        "-x",
        "--execute",
        nargs="?",
        const=True,
        help="Execute the dataflow. The argument is the variable name of a list of nodes; else execute all nodes.",
    )
    @argument(
        "-b",
        "--builder",
        help="Builder to which the module will be added and used for execution. Allows to pass Config and Adapters",
    )
    @argument(
        "-c",
        "--config",
        help="Config to build a Driver. Passing -c/--config at the same time as a Builder -b/--builder with a config will raise an exception.",
    )
    @argument(
        "-i",
        "--inputs",
        help="Execution inputs. The argument is the variable name of a dict of inputs; else {}.",
    )
    @argument(
        "-o",
        "--overrides",
        help="Execution overrides. The argument is the variable name of a dict of overrides; else {}.",
    )
    @argument(
        "--hide_results",
        action="store_true",
        help="Hides the automatic display of execution results.",
    )
    @argument(
        "-w",
        "--write_to_file",
        nargs="?",
        const=True,
        help="Write cell content to a file. The argument is the file path; else write to {module_name}.py",
    )
    @cell_magic
    def cell_to_module(self, line, cell):
        """Turn a notebook cell into a Hamilton module definition. This allows you to define
        and execute a dataflow from a single cell.

        For example:
        ```
        %%cell_to_module dataflow --display --execute
        def A() -> int:
          return 37

        def B(A: int) -> bool:
          return (A % 3) > 2
        ```
        """
        # shell.ex() is equivalent to exec(), but in the user namespace (i.e. notebook context).
        # This allows imports and functions defined in the magic cell %%cell_to_module to be
        # directly accessed from the notebook
        self.shell.ex(cell)

        args, unknown_args = parse_known_argstring(
            self.cell_to_module, line
        )  # specify how to parse by passing method
        self.resolve_unknown_args_cell_to_module(unknown_args)

        # validate variables exist in the user namespace expect `config` because it's a special case
        # will exit using `return` in case of error
        args_that_read_user_namespace = ["display", "builder", "final_vars", "inputs", "overrides"]
        for name, value in vars(args).items():
            if name not in args_that_read_user_namespace:
                continue

            # special case: args that can be passed as a flag (=True) without values
            if name in ["display", "final_vars"] and value is True:
                continue

            # main case: exit if variable is not in user namespace
            if value and self.shell.user_ns.get(value) is None:
                return f"KeyError: Received `--{name} {value}` but variable not found."

        # parse config; exit if config is invalid
        config = self.resolve_config_arg(args.config)
        if config is False:
            return

        # check if string instance because module_name has default `True`
        if isinstance(args.name, str) and isinstance(args.module_name, str):
            print(
                f"ValueError: Received both positional arg name={args.name} and named arg module_name={args.module_name}. Pass either one."
            )
            return

        # merged the positional arg `name` with named arg `module_name` for backwards compatibility
        module_name = args.module_name if isinstance(args.module_name, str) else args.name
        base_builder = self.shell.user_ns[args.builder] if args.builder else driver.Builder()
        inputs = self.shell.user_ns[args.inputs] if args.inputs else {}
        overrides = self.shell.user_ns[args.overrides] if args.overrides else {}
        display_config = (
            self.shell.user_ns[args.display] if args.display not in [True, None] else {}
        )

        # determine the Driver config
        # can't check from args.builder because it might be None
        if config and base_builder.config:
            return "AssertionError: Received a config -c/--config and a Builder -b/--builder with an existing config. Pass either one."

        # Decision: write to file before trying to build and execute Driver
        # See argument `help` for behavior details
        if args.write_to_file:
            if isinstance(args.write_to_file, str):
                file_path = Path(args.write_to_file)
            else:
                file_path = Path(f"{module_name}.py")
            file_path.write_text(cell)

        # create_module() is preferred over module_from_source() to simplify
        # the integration with the Hamilton UI which assumes physical Python modules
        try:
            cell_module = ad_hoc_utils.create_module(cell, module_name)
        except BaseException as e:
            print("Failed to build the module. Stack trace:")
            raise e
        self.shell.push({module_name: cell_module})

        # build the Driver. the Builder is copied to avoid conflict with the user namespace
        builder = base_builder.copy()
        dr = builder.with_config(config).with_modules(cell_module).build()

        # determine final vars
        if args.execute not in [True, None]:
            final_vars = self.shell.user_ns[args.execute]
        else:
            nodes = [n for n in dr.list_available_variables() if not n.is_external_input]
            final_vars = topological_sort(nodes)

        # visualize
        if args.display:
            # try/except `display_config` or inputs/overrides may be invalid
            try:
                if args.execute:
                    dot = dr.visualize_execution(
                        final_vars=final_vars,
                        inputs=inputs,
                        overrides=overrides,
                        **display_config,
                    )
                else:
                    dot = dr.display_all_functions(**display_config)
            except Exception as e:
                print(f"Failed to display {e}.\n\nThe display config was: {display_config}")
                dot = dr.display_all_functions()

            # handle output environment
            if self.notebook_env == "databricks":
                display_in_databricks(dot)
            else:
                display(dot)

        # execute
        if args.execute:
            results = dr.execute(
                final_vars=final_vars,
                inputs=inputs,
                overrides=overrides,
            )
            # normalize variable names that contain a `.` character like @pipe(step())
            results = {_normalize_result_names(name): value for name, value in results.items()}
            self.shell.push(results)

            if args.hide_results:
                return
            # results will follow the order of `final_vars` or topologically sorted if all vars
            display(*(results[n] for n in final_vars))

    # TODO unify the API and logic of `%%cell_to_module` and `%%incr_cell_to_module`
    @magic_arguments()
    @argument("module_name", nargs="?", help="Name for the module defined in this cell.")
    @argument(
        "identifier", type=int, help="Identifier for this cell w.r.t. the module being created."
    )
    @argument(
        "-c",
        "--config",
        help="Config to build a Driver. Passing -c/--config at the same time as a Builder -b/--builder with a config will raise an exception.",
    )
    @argument(
        "-b",
        "--builder",
        help="Builder to which the module will be added and used for execution. Allows to pass Config and Adapters",
    )
    @argument(
        "-d",
        "--display",
        nargs="?",
        const=True,
        help="Display the dataflow. The argument is the variable name of a dictionary of visualization kwargs; else {}.",
    )
    @argument(
        "-w",
        "--write_to_file",
        nargs="?",
        const=True,
        help="Write cell content to a file. The argument is the file path; else write to {module_name}.py",
    )
    @cell_magic
    def incr_cell_to_module(self, line, cell):
        """Incrementally build a module. This executes the cell and dynamically creates a Python module from its content.
        A Hamilton Driver is automatically instantiated with that module for variable `{MODULE_NAME}_dr`.

        > %%incr_cell_to_module MODULE_NAME -i IDENTIFIER --display
        """
        # This function mimics the logic of `.cell_to_module()`. Find more comments there.
        # Start by trying to execute the code cell.
        self.shell.ex(cell)

        # parse user inputs
        args, unknown_args = parse_known_argstring(self.incr_cell_to_module, line)
        self.resolve_unknown_args_cell_to_module(unknown_args)

        # check user inputs pointing to variables in user namespace
        args_that_read_user_namespace = ["display", "builder"]
        for name, value in vars(args).items():
            if name not in args_that_read_user_namespace:
                continue

            # special case: `display` can be passed as a flag (=True), without a config
            if name in ["display"] and value is True:
                continue

            # main case: exit if variable is not in user namespace
            if value and self.shell.user_ns.get(value) is None:
                return f"KeyError: Received `--{name} {value}` but variable not found."

        # TODO convert -i to -id
        if args.identifier is None:
            raise ValueError("`-id/--identifier` is required. Please provide an id for this cell.")

        # parse config; exit if config is invalid
        config = self.resolve_config_arg(args.config)
        if config is False:
            return

        # set parsed arguments
        module_name = args.module_name
        base_builder = self.shell.user_ns[args.builder] if args.builder else driver.Builder()
        display_config = (
            self.shell.user_ns[args.display] if args.display not in [True, None] else {}
        )

        # determine the Driver config
        # can't check from args.builder because it might be None
        if config and base_builder.config:
            return "AssertionError: Received a config -c/--config and a Builder -b/--builder with an existing config. Pass either one."

        # store current cell in state
        self.incremental_cells_state[module_name][args.identifier] = cell

        # build module source from multiple cells
        module_dict = self.incremental_cells_state[module_name]
        sorted_module_keys = sorted(list(module_dict.keys()))
        module_source = "\n\n".join([module_dict[k] for k in sorted_module_keys])
        multi_cell_module = ad_hoc_utils.create_module(module_source, module_name)
        self.shell.push({module_name: multi_cell_module})

        # Decision: write to file before trying to build and execute Driver
        # See argument `help` for behavior details
        if args.write_to_file:
            if isinstance(args.write_to_file, str):
                file_path = Path(args.write_to_file)
            else:
                file_path = Path(f"{module_name}.py")
            file_path.write_text(module_source)

        # build Driver
        builder = base_builder.copy()
        dr = builder.with_config(config).with_modules(multi_cell_module).build()

        # visualize
        if args.display:
            # try/except `display_config` or inputs/overrides may be invalid
            try:
                dot = dr.display_all_functions(**display_config)
            except Exception as e:
                print(f"Failed to display {e}.\n\nThe display config was: {display_config}")
                dot = dr.display_all_functions()

            # handle output environment
            if self.notebook_env == "databricks":
                display_in_databricks(dot)
            else:
                display(dot)

    @magic_arguments()  # needed on top to enable parsing
    @argument("module_name", help="Module name to print.")  # required argument
    @line_magic
    def print_module(self, line):
        """Prints the contents of a dynamic module we've been creating."""
        args = parse_argstring(self.incr_cell_to_module, line)
        if args.module_name in self.incremental_cells_state:
            module_dict = self.incremental_cells_state[args.module_name]
            sorted_module_keys = sorted(list(module_dict[args.module_name].keys()))
            module_source = "\n\n".join([module_dict[k] for k in sorted_module_keys])
            display(Code(module_source))
        else:
            print(f"KeyError: `{args.module_name}` not found.")

    @magic_arguments()  # needed on top to enable parsing
    @argument("module_name", help="Module to print.")  # required argument
    @line_magic
    def reset_module(self, line):
        args = parse_argstring(self.incr_cell_to_module, line)
        if args.module_name in self.incremental_cells_state:
            del self.incremental_cells_state[args.module_name]
            print(f"Reset `{args.module_name}`")
        else:
            print(f"KeyError: `{args.module_name}` not found.")

    @magic_arguments()
    @argument("name", type=str, help="Creates a dictionary fromt the cell's content.")
    @cell_magic
    def set_dict(self, line: str, cell: str):
        """Execute the cell and store all assigned variables as inputs"""
        args = parse_argstring(self.set_dict, line)
        self.shell.user_ns[args.name] = execute_and_get_assigned_values(self.shell, cell)

    @line_magic
    def insert_module(self, line):
        """Alias for `%module_to_cell`."""
        self.module_to_cell(line)

    @line_magic
    def module_to_cell(self, line):
        """Insert in the next cell the source code from the module (.py)
        at the path specified by `line`.

        This is useful if you have an existing Hamilton module and want to pull in the contents
        for further development in the notebook.
        """

        # JavaScript magic to generate a new cell
        insert_cell_with_content()

        module_path = Path(line)
        # insert our custom %%with_functions magic at the top of the cell
        header = f"%%cell_to_module {module_path.stem}\n"
        module_source = module_path.read_text()
        # insert source code as text in the next cell
        self.shell.set_next_input(header + module_source, replace=False)


def load_ipython_extension(ipython: InteractiveShellApp):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    ipython.register_magics(HamiltonMagics)



from typing import Any, Dict, Optional, Type, Union

import narwhals as nw

from hamilton.lifecycle import api


class NarwhalsAdapter(api.NodeExecutionMethod):
    """Adapter to make it simpler to use narwhals with Hamilton.

    .. code-block:: python

            from hamilton import base, driver
            from hamilton.plugins import h_narwhals
            import example

            # pandas
            dr = (
                driver.Builder()
                .with_config({"load": "pandas"})
                .with_modules(example)
                .with_adapters(
                    h_narwhals.NarwhalsAdapter(),
                    h_narwhals.NarwhalsDataFrameResultBuilder(
                        base.PandasDataFrameResult()
                    ),
                )
                .build()
            )
            result = dr.execute(
                 [example.group_by_mean, example.example1],
                 inputs={"col_name": "a"}
            )

    """

    def run_to_execute_node(
        self,
        *,
        node_name: str,
        node_tags: Dict[str, Any],
        node_callable: Any,
        node_kwargs: Dict[str, Any],
        task_id: Optional[str],
        **future_kwargs: Any,
    ) -> Any:
        """This method is responsible for executing the node and returning the result.

        It uses `nw_kwargs` from the node tags to know if any special flags should be passed to the narwhals
        decorator function.

        :param node_name: Name of the node.
        :param node_tags: Tags of the node.
        :param node_callable: Callable of the node.
        :param node_kwargs: Keyword arguments to pass to the node.
        :param task_id: The ID of the task, none if not in a task-based environment
        :param future_kwargs: Additional keyword arguments -- this is kept for backwards compatibility
        :return: The result of the node execution -- up to you to return this.
        """
        nw_kwargs = {}
        if "nw_kwargs" in node_tags:
            nw_kwargs = {k: True for k in node_tags["nw_kwargs"]}
        nw_func = nw.narwhalify(node_callable, **nw_kwargs)
        return nw_func(**node_kwargs)


class NarwhalsDataFrameResultBuilder(api.ResultBuilder):
    """Builds the result. It unwraps the narwhals parts of it and delegates to the passed in result builder.

    .. code-block:: python

        from hamilton import base, driver
        from hamilton.plugins import h_narwhals, h_polars
        import example

        # polars
        dr = (
            driver.Builder()
            .with_config({"load": "polars"})
            .with_modules(example)
            .with_adapters(
                h_narwhals.NarwhalsAdapter(),
                h_narwhals.NarwhalsDataFrameResultBuilder(
                   h_polars.PolarsDataFrameResult()
                ),
            )
            .build()
        )
        result = dr.execute(
            ["group_by_mean", "example1"],
            inputs={"col_name": "a"}
        )
    """

    def __init__(self, result_builder: Union[api.ResultBuilder, api.LegacyResultMixin]):
        self.result_builder = result_builder

    def build_result(self, **outputs: Any) -> Any:
        """Given a set of outputs, build the result.

        :param outputs: the outputs from the execution of the graph.
        :return: the result of the execution of the graph.
        """
        de_narwhaled_outputs = {}
        for key, value in outputs.items():
            if isinstance(value, (nw.DataFrame, nw.Series)):
                de_narwhaled_outputs[key] = nw.to_native(value)
            else:
                de_narwhaled_outputs[key] = value

        return self.result_builder.build_result(**de_narwhaled_outputs)

    def output_type(self) -> Type:
        """Returns the output type of this result builder
        :return: the type that this creates
        """
        return self.result_builder.output_type()



import pandas as pd
import pyarrow
import pytest

from hamilton.plugins import h_pyarrow


@pytest.fixture()
def pandas():
    return pd.DataFrame({"a": [0, 1, 2], "b": ["a", "b", "c"]})


def test_pandas_to_pyarrow(pandas):
    result_builder = h_pyarrow.PyarrowTableResult()
    data = {"df": pandas}
    # ResultBuilder receive unpacked dict as arg, i.e., kwargs only
    table = result_builder.build_result(**data)
    assert isinstance(table, pyarrow.Table)


def test_fail_for_multiple_outputs(pandas):
    result_builder = h_pyarrow.PyarrowTableResult()
    data = {"df": pandas, "df2": pandas}
    with pytest.raises(AssertionError):
        result_builder.build_result(**data)



import inspect

import pandas as pd
from kedro.pipeline import node

from hamilton.plugins import h_kedro


def test_parse_k_node_str_output():
    def preprocess_companies(companies: pd.DataFrame) -> pd.DataFrame:
        """Preprocesses the data for companies."""
        companies["iata_approved"] = companies["iata_approved"].astype("category")
        return companies

    kedro_node = node(
        func=preprocess_companies,
        inputs="companies",
        outputs="preprocessed_companies",
        name="preprocess_companies_node",
    )
    h_nodes = h_kedro.k_node_to_h_nodes(kedro_node)
    assert len(h_nodes) == 1
    assert h_nodes[0].name == "preprocessed_companies"
    assert h_nodes[0].type == inspect.signature(preprocess_companies).return_annotation


def test_parse_k_node_list_outputs():
    def multi_outputs() -> dict:
        return dict(a=1, b=2)

    kedro_node = node(
        func=multi_outputs,
        inputs=None,
        outputs=["a", "b"],
    )
    h_nodes = h_kedro.k_node_to_h_nodes(kedro_node)
    node_names = [n.name for n in h_nodes]
    assert len(h_nodes) == 3
    assert "multi_outputs" in node_names
    assert "a" in node_names
    assert "b" in node_names


def test_parse_k_node_dict_outputs():
    def multi_outputs() -> dict:
        return dict(a=1, b=2)

    kedro_node = node(
        func=multi_outputs,
        inputs=None,
        outputs={"a": "a", "b": "b"},
    )
    h_nodes = h_kedro.k_node_to_h_nodes(kedro_node)
    node_names = [n.name for n in h_nodes]
    assert len(h_nodes) == 3
    assert "multi_outputs" in node_names
    assert "a" in node_names
    assert "b" in node_names



import abc
import dataclasses
import enum
from collections import defaultdict
from typing import Any, Collection, Dict, List, Optional, Set, Tuple

from hamilton import node
from hamilton.execution import graph_functions
from hamilton.execution.graph_functions import get_node_levels, topologically_sort_nodes
from hamilton.lifecycle import base as lifecycle_base
from hamilton.node import Node, NodeType

"""General utilities for grouping nodes in a DAG"""


class NodeGroupPurpose(enum.Enum):
    """Node groups have a purpose. The standard type is "EXECUTE_BLOCK",
    which just executes a subdag. The other types are for creating "expanders",
    which can dictate how subsequent nodes expand into a subdag.

    """

    EXPAND_UNORDERED = "expand_unordered"  # DAG that ends in a parameter for the next subdag to
    GATHER = "gather"  # DAG that begins in the collection of a parallel block
    EXECUTE_BLOCK = "execute_block"  # DAG that is a standard block of nodes
    EXECUTE_SINGLE = "execute_single"  # DAG that is just a single node

    def is_expander(self):
        return self.value in ["expand_unordered"]

    def is_gatherer(self):
        return self.value in ["gather"]


@dataclasses.dataclass
class NodeGroup:
    """Represents a simple grouping of nodes into a single unit.
    This has some utility functions to determine inputs, outputs, etc...
    Note that this has the following properties:
    1. ID is unique across all node groups
    2. Nodes are unique across all node groups (one node to group)
    """

    base_id: str  # Unique ID for node group.
    spawning_task_base_id: Optional[str]
    nodes: List[Node]
    purpose: NodeGroupPurpose  # TODO -- derive this (or not?)
    # set of available nodes by this task for querying
    available_nodes: Set[str] = dataclasses.field(init=False)

    def __post_init__(self):
        self.available_nodes = {node_.name for node_ in self.nodes}

    def __hash__(self):
        return hash(self.base_id)

    def __eq__(self, other):
        return self.base_id == other.base_id

    def get_expander_node(self) -> Optional[Node]:
        """Returns the expander node for this node group, if it exists"""
        candidates = [n for n in self.nodes if n.node_role == NodeType.EXPAND]
        if candidates:
            return candidates[0]
        return None

    def get_collector_node(self) -> Optional[Node]:
        """Returns the collector node for this node group, if it exists"""
        candidates = [n for n in self.nodes if n.node_role == NodeType.COLLECT]
        if candidates:
            return candidates[0]
        return None

    def produces(self, node_name: str) -> bool:
        return node_name in self.available_nodes


@dataclasses.dataclass
class TaskSpec(NodeGroup):
    """Represents the spec for a task. This has no actual input/output data associated with it,
    but has information needed to know how one might execute.

    You can think of this as a "template" for a task, where the inputs are not yet known.
    Furthermore, there could be multiple instances of this task spec, each with different inputs.
    """

    outputs_to_compute: Collection[str]  # list of output names to compute
    overrides: Dict[str, Any]  # overrides for the task, fixed at the time of creation
    adapter: lifecycle_base.LifecycleAdapterSet
    base_dependencies: List[str]  # list of tasks that must be completed before this task can run

    def get_input_vars(self) -> Tuple[List[str], List[str]]:
        """Returns the node-level dependencies for this node group.
        This is all of the sources in the subdag.

        :return: A tuple consisting of (a) the list of required inputs variables and (b)
        the list of optional input variables
        """
        all_node_names = {n.name for n in self.nodes}
        required_variables = set()
        optional_variables = set()
        for node_ in self.nodes:
            if node_.name in self.overrides:
                continue
            if node_.user_defined:
                # if the node is user-defined then that
                # means we need it for the execution
                required_variables.add(node_.name)
            for dependency in node_.dependencies:
                if dependency.name not in self.overrides:
                    if dependency.user_defined or dependency.name not in all_node_names:
                        if node_.requires(dependency.name):
                            if dependency.name in optional_variables:
                                optional_variables.remove(dependency.name)
                            required_variables.add(dependency.name)
                        else:
                            optional_variables.add(dependency.name)

        return list(required_variables), list(optional_variables)


@dataclasses.dataclass
class TaskImplementation(TaskSpec):
    """Represents the "execution ready" spec for a task. This contains two essential pieces of information:

    1. The inputs/outputs
    2. The "spawning" task that spawned this

    Thus the task_id is unique (as it contains the spawning task's ID + the group ID), but the base_id is not, as
    we could have multiple versions of this task running in parallel.
    """

    # task whose result spawned these tasks
    # If this is none, it means the graph itself spawned these tasks
    group_id: Optional[str]
    realized_dependencies: Dict[str, List[str]]  # realized dependencies are the actual dependencies
    # Note that these are lists as we have "gather" operations
    spawning_task_id: Optional[str]  # task that spawned this task
    task_id: str = dataclasses.field(init=False)
    dynamic_inputs: Dict[str, Any] = dataclasses.field(default_factory=dict)
    run_id: str = dataclasses.field(default_factory=str)

    def bind(self, dynamic_inputs: Dict[str, Any]) -> "TaskImplementation":
        """Binds dynamic inputs to the task spec, returning a new task spec"""
        return dataclasses.replace(self, dynamic_inputs={**dynamic_inputs, **self.dynamic_inputs})

    @staticmethod
    def determine_task_id(base_id: str, spawning_task: Optional[str], group_id: Optional[str]):
        return ".".join(
            filter(lambda i: i is not None, [spawning_task, group_id, base_id])
        )  # This will do for now...

    def __post_init__(self):
        super(TaskImplementation, self).__post_init__()
        self.task_id = self.determine_task_id(self.base_id, self.spawning_task_id, self.group_id)


class GroupingStrategy(abc.ABC):
    """Base class for grouping nodes"""

    @abc.abstractmethod
    def group_nodes(self, nodes: List[node.Node]) -> List[NodeGroup]:
        """Groups nodes into a list of node groups"""
        pass


class GroupByRepeatableBlocks(GroupingStrategy):
    """Groups nodes by repeatable blocks. For every set of nodes betweeen
    a "parallel[]" and a "collect[]", we place them in a "block", which is its own group.
    Note that this is specifically built to allow parallel/remote execution of blocks, with local
    execution of non-parallel groups (that might generate/read from the blocks).
    """

    @staticmethod
    def nodes_after_last_expand_block(
        collect_node: node.Node,
    ) -> Tuple[node.Node, List[node.Node]]:
        """Utility function to yield all nodes between a start and an end node.
        This returns all nodes for which the following conditions are met:

        1. It contains the start node as an ancestor
        2. It contains a node that matches the terminate condition as a dependent

        :param collect_node: Collector node to trace from
        :return: The nodes in between start and end
        """

        def is_expander(node_: node.Node) -> bool:
            return node_.node_role in [NodeType.EXPAND]

        return graph_functions.nodes_between(collect_node, is_expander)

    def group_nodes(self, nodes: List[node.Node]) -> List[NodeGroup]:
        """Groups nodes into blocks. This works as follows:
        1. Fina all the Parallelizable[] nodes in the DAG
        2. For each of those, do a DFS until the next Collect[] node
        3. Create a group for each of those
        4. Add all other nodes to their own group


        :param nodes:
        :return:
        """
        collectors = [node_ for node_ in nodes if node_.node_role == NodeType.COLLECT]
        groups = []
        visited = set()
        # We have to do this as we may have overrides to handle
        node_names = {node_.name for node_ in nodes}
        for collector in collectors:
            expander, nodes_in_block = self.nodes_after_last_expand_block(collector)
            # TODO -- add error message for conflicting groups...
            # if expander in visited:
            #     raise ValueError(f"Multiple collect nodes cannot trace "
            #                      f"back to the same expander. ")
            expander_name = f"expand-{expander.name}"
            if expander.name in node_names:
                groups.append(
                    NodeGroup(
                        base_id=f"expand-{expander.name}",
                        spawning_task_base_id=None,
                        nodes=[expander],
                        purpose=NodeGroupPurpose.EXPAND_UNORDERED,
                    )
                )
                # In thie case of a strange override, we may end up with this
                # breaking, as a node in the block could be missing
                # This is an undefined case, but we'll likely end up with an error
                groups.append(
                    NodeGroup(
                        base_id=f"block-{expander.name}",
                        spawning_task_base_id=expander_name,
                        nodes=nodes_in_block,
                        purpose=NodeGroupPurpose.EXECUTE_BLOCK,
                    )
                )
                visited.update(nodes_in_block + [expander])
            groups.append(
                NodeGroup(
                    base_id=f"collect-{expander.name}",
                    spawning_task_base_id=expander_name,
                    nodes=[collector],
                    purpose=NodeGroupPurpose.GATHER,
                )
            )
            visited.update([collector])

        remaining_nodes = [node_ for node_ in nodes if node_ not in visited]
        for node_ in remaining_nodes:
            groups.append(
                NodeGroup(
                    base_id=node_.name,
                    spawning_task_base_id=None,  # Nothing spawns this task
                    nodes=[node_],
                    purpose=NodeGroupPurpose.EXECUTE_SINGLE,
                )
            )
        return groups


def convert_node_type_to_group_purpose(node_type: NodeType) -> NodeGroupPurpose:
    """Utility type to map node type to group purpose. Note its not 1:1, but its pretty close."""
    if node_type == NodeType.EXPAND:
        return NodeGroupPurpose.EXPAND_UNORDERED
    if node_type == NodeType.COLLECT:
        return NodeGroupPurpose.GATHER
    return NodeGroupPurpose.EXECUTE_BLOCK


class GroupNodesIndividually(GroupingStrategy):
    """Groups nodes into individual blocks."""

    def group_nodes(self, nodes: List[node.Node]):
        return [
            NodeGroup(
                base_id=node_.name,
                nodes=[node_],
                purpose=convert_node_type_to_group_purpose(node_.node_role),
                spawning_task_base_id=None,
            )
            for node_ in nodes
        ]


class GroupNodesAllAsOne(GroupingStrategy):
    """Groups nodes all into one block. TODO -- add validation."""

    def group_nodes(self, nodes: List[node.Node]):
        return [
            NodeGroup(
                base_id="root",
                spawning_task_base_id=None,
                nodes=nodes,
                purpose=convert_node_type_to_group_purpose(nodes[0].node_role),
            )
        ]


class GroupNodesByLevel(GroupingStrategy):
    def group_nodes(self, nodes: List[node.Node]) -> List[NodeGroup]:
        in_order = topologically_sort_nodes(nodes)
        node_levels = get_node_levels(in_order)
        nodes_by_level = defaultdict(list)
        nodes_by_name = {node_.name: node_ for node_ in nodes}
        for node_name, level in node_levels.items():
            nodes_by_level[level].append(nodes_by_name[node_name])
        out = []
        for i in range(len(nodes_by_level)):
            nodes = nodes_by_level[i]
            out.append(
                NodeGroup(
                    base_id=f"level_{i}",
                    spawning_task_base_id=None,
                    nodes=nodes,
                    purpose=NodeGroupPurpose.EXECUTE_BLOCK,
                )
            )
        return out


def create_task_plan(
    node_groups: List[NodeGroup],
    outputs: List[str],
    overrides: Dict[str, Any],
    adapter: lifecycle_base.LifecycleAdapterSet,
) -> List[TaskSpec]:
    """Creates tasks from node groups. This occurs after we group and after execute() is called in
    the driver. Knowing what the user wants, we can finally create the tasks.

    This does the following:

    1. Determines any tasks that are not in the critical path, deletes them
    2. For every task in the critical path
        - Sets the outputs required from that task (all user-specified outputs that reside in that task)
        - Instantiates and returns the task

    :param node_groups: Groups of nodes to form tasks
    :param outputs: output nodes the user wants
    :param overrides: Overrides that short-circuit the execution
    :param adapter: LifecycleAdapterSet to use for execution
    :return: A list of task specs that we will execute later
    """

    outputs_set = set(outputs)
    out = []
    node_to_task_map = {}
    for node_group in node_groups:
        outputs = []
        nodes_in_group = set([node_.name for node_ in node_group.nodes])
        for node_ in node_group.nodes:
            # If this is overridden we can skip it
            # Note there may be a strange bug with grouping/overrides (E.G. downstream of overrides)
            if node_.name in overrides:
                continue
            node_to_task_map[node_.name] = node_group
            is_output = False
            if node_.name in outputs_set:
                is_output = True
            for depended_on_by in node_.depended_on_by:
                if (
                    depended_on_by.name not in overrides
                    and depended_on_by.name not in nodes_in_group
                ):
                    is_output = True
                    break
            if is_output:
                outputs.append(node_.name)

        task_spec = TaskSpec(
            base_id=node_group.base_id,
            spawning_task_base_id=node_group.spawning_task_base_id,
            nodes=node_group.nodes,
            purpose=node_group.purpose,
            outputs_to_compute=outputs,
            adapter=adapter,
            overrides={
                node_.name: overrides[node_.name]
                for node_ in node_group.nodes
                if node_.name in overrides
            },
            base_dependencies=[],
        )
        out.append(task_spec)
    # Now we need to go through and add dependencies to each task
    # We should be able to do this in the single pass if we guarenteed topological order,
    # but for now we're doing this in the second pass
    for task_spec in out:
        task_dependencies = set()
        nodes_in_group = set([node_.name for node_ in task_spec.nodes])
        for node_ in task_spec.nodes:
            if node_.name in overrides:
                continue
            for dependency in node_.dependencies:
                if dependency.name not in overrides and dependency.name not in nodes_in_group:
                    task_containing_dependency = node_to_task_map.get(dependency.name)
                    # If its optional, we don't need to add it as a dependency
                    if task_containing_dependency is None and not node_.requires(dependency.name):
                        continue
                    if task_containing_dependency is None:
                        raise ValueError(
                            f"Dependency is None and requires {dependency.name}."
                            f"We're somehow in a bad state. Please reach out on slack."
                        )
                    task_dependencies.add(task_containing_dependency.base_id)
        task_spec.base_dependencies = list(task_dependencies)
    return out



import pathlib
from typing import List, Type

import pytest
import yaml

from hamilton.function_modifiers.adapters import resolve_adapter_class
from hamilton.io.data_adapters import DataLoader, DataSaver
from hamilton.plugins.yaml_extensions import PrimitiveTypes, YAMLDataLoader, YAMLDataSaver

TEST_DATA_FOR_YAML = [
    (1, "int.yaml"),
    ("string", "string.yaml"),
    (True, "bool.yaml"),
    ({"key": "value"}, "test.yaml"),
    ([1, 2, 3], "data.yaml"),
    ({"nested": {"a": 1, "b": 2}}, "config.yaml"),
]


@pytest.mark.parametrize("data, file_name", TEST_DATA_FOR_YAML)
def test_yaml_loader(tmp_path: pathlib.Path, data, file_name):
    path = tmp_path / pathlib.Path(file_name)
    with path.open(mode="w") as f:
        yaml.dump(data, f)
    assert path.exists()
    loader = YAMLDataLoader(path)
    loaded_data = loader.load_data(type(data))
    assert loaded_data[0] == data


@pytest.mark.parametrize("data, file_name", TEST_DATA_FOR_YAML)
def test_yaml_saver(tmp_path: pathlib.Path, data, file_name):
    path = tmp_path / pathlib.Path(file_name)
    saver = YAMLDataSaver(path)
    saver.save_data(data)
    assert path.exists()
    with path.open("r") as f:
        loaded_data = yaml.safe_load(f)
    assert data == loaded_data


@pytest.mark.parametrize("data, file_name", TEST_DATA_FOR_YAML)
def test_yaml_loader_and_saver(tmp_path: pathlib.Path, data, file_name):
    path = tmp_path / pathlib.Path(file_name)
    saver = YAMLDataSaver(path)
    saver.save_data(data)
    assert path.exists()
    loader = YAMLDataLoader(path)
    loaded_data = loader.load_data(type(data))
    assert data == loaded_data[0]


@pytest.mark.parametrize(
    "type_,classes,correct_class",
    [(t, [YAMLDataLoader], YAMLDataLoader) for t in PrimitiveTypes],
)
def test_resolve_correct_loader_class(
    type_: Type[Type], classes: List[Type[DataLoader]], correct_class: Type[DataLoader]
):
    assert resolve_adapter_class(type_, classes) == correct_class


@pytest.mark.parametrize(
    "type_,classes,correct_class",
    [(t, [YAMLDataSaver], YAMLDataSaver) for t in PrimitiveTypes],
)
def test_resolve_correct_saver_class(
    type_: Type[Type], classes: List[Type[DataSaver]], correct_class: Type[DataLoader]
):
    assert resolve_adapter_class(type_, classes) == correct_class



import pandas as pd


def avg_3wk_spend(spend: pd.Series) -> pd.Series:
    """Rolling 3 week average spend."""
    return spend.rolling(3).mean()


def spend_per_signup(spend: pd.Series, signups: pd.Series) -> pd.Series:
    """The cost per signup in relation to spend."""
    return spend / signups



import pathlib

import lancedb
import numpy as np
from datasets import Dataset, DatasetDict

from hamilton.plugins import huggingface_extensions


def test_hfds_loader():
    path_to_test = "tests/resources/hf_datasets"
    reader = huggingface_extensions.HuggingFaceDSLoader(path_to_test)
    ds, metadata = reader.load_data(DatasetDict)

    assert huggingface_extensions.HuggingFaceDSLoader.applicable_types() == list(
        huggingface_extensions.HF_types
    )
    assert reader.applies_to(DatasetDict)
    assert reader.applies_to(Dataset)
    assert ds.shape == {"train": (1, 3)}


def test_hfds_parquet_saver(tmp_path: pathlib.Path):
    file_path = tmp_path / "testhf.parquet"
    saver = huggingface_extensions.HuggingFaceDSParquetSaver(file_path)
    ds = Dataset.from_dict({"a": [1, 2, 3]})
    metadata = saver.save_data(ds)
    assert file_path.exists()
    assert metadata["dataset_metadata"] == {
        "columns": ["a"],
        "features": {"a": {"_type": "Value", "dtype": "int64"}},
        "rows": 3,
        "size_in_bytes": None,
    }
    assert "file_metadata" in metadata
    assert huggingface_extensions.HuggingFaceDSParquetSaver.applicable_types() == list(
        huggingface_extensions.HF_types
    )
    assert saver.applies_to(DatasetDict)
    assert saver.applies_to(Dataset)


def test_hfds_lancedb_saver(tmp_path: pathlib.Path):
    db_client = lancedb.connect(tmp_path / "lancedb")
    saver = huggingface_extensions.HuggingFaceDSLanceDBSaver(db_client, "test_table")
    ds = Dataset.from_dict({"vector": [np.array([1.0, 2.0, 3.0])], "named_entities": ["a"]})
    metadata = saver.save_data(ds)
    assert metadata == {
        "dataset_metadata": {
            "columns": ["vector", "named_entities"],
            "features": {
                "named_entities": {"_type": "Value", "dtype": "string"},
                "vector": {"_type": "Sequence", "feature": {"_type": "Value", "dtype": "float64"}},
            },
            "rows": 1,
            "size_in_bytes": None,
        },
        "db_meta": {"table_name": "test_table"},
    }
    assert db_client.open_table("test_table").search().to_list() == [
        {"named_entities": "a", "vector": [1.0, 2.0, 3.0]}
    ]



import dask

from hamilton import telemetry

# disable telemetry for all tests!
telemetry.disable_telemetry()

# required until we fix the DataFrameResultBuilder to work with dask-expr
dask.config.set({"dataframe.query-planning": False})



import pandas as pd

from hamilton.function_modifiers import does, extract_columns, extract_fields, tag


def _sum_multiply(param0: int, param1: int, param2: int) -> pd.DataFrame:
    return pd.DataFrame([{"param0a": param0, "param1b": param1, "param2c": param2}])


def _sum(param0: int, param1: int, param2: int) -> dict:
    return {"total": param0 + param1 + param2}


@extract_columns("param1b")
@does(_sum_multiply)
def to_modify(param0: int, param1: int, param2: int = 2) -> pd.DataFrame:
    """This is a dummy function showing extract_columns with does."""
    pass


@extract_fields({"total": int})
@does(_sum)
def to_modify_2(param0: int, param1: int, param2: int = 2) -> dict:
    """This is a dummy function showing extract_fields with does."""
    pass


def _dummy(**values) -> dict:
    return {f"out_{k.split('_')[1]}": v for k, v in values.items()}


@extract_fields({"out_value1": int, "out_value2": str})
@tag(test_key="test-value")
# @check_output(data_type=dict, importance="fail")  To fix see https://github.com/dagworks-inc/hamilton/issues/249
@does(_dummy)
def uber_decorated_function(in_value1: int, in_value2: str) -> dict:
    pass
